/* ==========================================================================
 * mod_auth_openldap - Apache module for LDAP Authentication using OpenLDAP
 * --------------------------------------------------------------------------
 * Copyright (c) 2005  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include <stdio.h>	/* snprintf(3) */
#include <stdlib.h>	/* malloc(3) free(3) */
#include <stdarg.h>	/* va_list va_start va_end */

#include <string.h>	/* strcasecmp(3) strsep(3) strerror(3) */

#include <errno.h>

#include <assert.h>

#include <ldap.h>	/* ldap_open() */


static const struct auth_config {
	int enabled;
	int require_ssl;
	char *host;
	char *port;
	char *version;
	char *tls_mode;
	char *binddn;
	char *password;
	char *basedn;
	char *scope;
	char *filter;
} auth_config_initializer = {
	.enabled	= 0,
	.require_ssl	= 1,
	.host		= "127.0.0.1",
	.port		= "389",
	.version	= "3",
	.tls_mode	= "StartTLS",
	.scope		= "one",
	.filter		= "(uid=%u)"
};


/*
 * Apache module parameter declaration
 */
module MODULE_VAR_EXPORT openldap_auth_module;


/*
 * pool *p:	apache pool allocator
 * char *dn:	directory name
 */
static void *create_dir_config(pool *p, char *dn) {
	struct auth_config *cfg;

	cfg	= ap_pcalloc(p, sizeof *cfg);

	assert(cfg);

	*cfg	= auth_config_initializer;

	return cfg;
} /* create_dir_config */


static char *create_filter(pool *p, struct auth_config *cfg, char *user) {
	static const int growby	= 128;
	char *buf		= 0;
	size_t bufsiz		= 0;
	char *tmp;
	struct { char *pos, *end; } dst;
	struct { char *pos; } src;

growbuf:
	if (!(tmp = realloc(buf,bufsiz + growby)))
		return free(buf), (char *)0;

	buf	= tmp;
	bufsiz	+= growby;

	dst.pos	= buf;
	dst.end	= buf + bufsiz;

	src.pos	= cfg->filter;

	while (src.pos && *src.pos) {
		if (*src.pos == '%') {
			tmp	= src.pos;

			switch (*(src.pos + 1)) {
			case 'u':
				src.pos	= user;

				break;
			default:
				goto copyout;
			}

			while (src.pos && *src.pos) {
				if (isalnum((unsigned char)*src.pos)) {
					if (dst.pos < dst.end) {
						*(dst.pos++)	= *(src.pos++);
					} else
						goto growbuf;
				} else {
					static const char map[] = {
						'0', '1', '2', '3',
						'4', '5', '6', '7',
						'8', '9', 'a', 'b',
						'c', 'd', 'e', 'f'
					};

					if (!(dst.end - dst.pos > 3))
						goto growbuf;

					*(dst.pos++)	= '\\';
					*(dst.pos++)	= map[(((unsigned char)*src.pos >> 4) & 0x0f)];
					*(dst.pos++)	= map[(((unsigned char)*src.pos) & 0x0f)];

					src.pos++;
				}
			}

			src.pos	= tmp + 2;
		} else {
copyout:
			if (dst.pos < dst.end) {
				*(dst.pos++)	= *(src.pos++);
			} else
				goto growbuf;
		}
	}

	if (!bufsiz || dst.pos == buf)
		return free(buf), (char *)0;

	tmp	= ap_pstrndup(p,buf,dst.pos - buf);

	free(buf);

	return tmp;
#if 0
	if (dst.pos >= dst.end)
		goto growbuf;

	*dst.pos	= '\0';

	return buf;
#endif
} /* create_filter() */


static char *nulstr(const char *str) {
	return (char *)((str)? str : "");
} /* nulstr() */


static void split(int *argc, char *argv[], int len, char *line, char *sep) {
	char **ap;
	char *nxt	= line;

	if (!sep)
		sep	= " \t";

	for (*argc = 0, ap = argv; ap < &argv[len - 1];) {
		if ((*ap = strsep(&nxt,sep)) == 0)
			break;

		if (**ap != '\0')
			ap++, (*argc)++;
	}
	*ap	= 0;
} /* split() */


static void ldapauth_warn(request_rec *r, const char *fmt, ...) {
	va_list ap;
	char msg[1024];
	int len;

	va_start(ap,fmt);

	len	= vsnprintf(msg,sizeof msg,fmt,ap);

	if (len > 0)
		ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_ERR,r->server,"[AuthLDAP] access to %s failed for %s, reason %s",r->uri,ap_get_remote_host(r->connection,r->per_dir_config,REMOTE_NAME),msg);

	va_end(ap);

	return /* void */;
} /* ldapauth_warn() */


static void ldapauth_debug(request_rec *r, const char *fmt, ...) {
	va_list ap;
	char msg[1024];
	int len;

	va_start(ap,fmt);

	len	= vsnprintf(msg,sizeof msg,fmt,ap);

	if (len > 0)
		ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_DEBUG,r->server,"[AuthLDAP] access to %s for %s said %s",r->uri,ap_get_remote_host(r->connection,r->per_dir_config,REMOTE_NAME),msg);

	va_end(ap);

	return /* void */;
} /* ldapauth_debug() */


static struct ldap *ldapauth_open(request_rec *r, struct auth_config *cfg) {
	LDAP *ld		= 0;
	const char *errstr	= 0;
	int version;

	version	= strtonum(cfg->version,1,3,&errstr);

	if (errstr) {
		ldapauth_warn(r,"Invalid protocol version: %s",cfg->version);
		
		return 0;
	}

	if (0 == strcasecmp(cfg->tls_mode,"LDAPS")) {
		char *hosts[8]	= { 0 };
		char *tmp	= ap_pstrdup(r->pool,cfg->host);
		int nhosts, i, ret;

		ldapauth_debug(r,"@line %d",__LINE__);

		if (!tmp) {
			ldapauth_warn(r,"Internal error: %s",strerror(errno));

			return 0;
		}

		split(&nhosts,hosts,sizeof hosts / sizeof *hosts,tmp," \t:");

		if (!nhosts) {
			ldapauth_warn(r,"No hosts provided");

			return 0;
		}

		for (i = 0; i < nhosts; i++) {
			char *url	= ap_pstrcat(r->pool,"ldaps://",hosts[i],":",cfg->port,(char *)0);

			if (!url) {
				ldapauth_warn(r,"Internal error: %s",strerror(errno));

				return 0;
			}

			ldapauth_debug(r,"@line %d",__LINE__);

			ret	= ldap_initialize(&ld,url);

			if (ret != LDAP_SUCCESS) {
				ld	= 0;

				ldapauth_warn(r,"Failed to connect to %s: %s",url,ldap_err2string(ret));
			} else {
				assert(ld);

				break;
			}
		}

		if (ld && LDAP_OPT_SUCCESS != ldap_set_option(ld,LDAP_OPT_PROTOCOL_VERSION,&version)) {
			ldapauth_warn(r,"Failed to set protocol version to %s",cfg->version);

			ldap_unbind_s(ld);

			ld	= 0;
		}

		ldapauth_debug(r,"@line %d",__LINE__);

	} else if (0 == strcasecmp(cfg->tls_mode,"StartTLS") || 0 == strcasecmp(cfg->tls_mode,"Plain")) {
		long port		= strtonum(cfg->port,1,65535,&errstr);

		ldapauth_debug(r,"@line %d",__LINE__);

		if (errstr)
			ldapauth_warn(r,"Invalid port: %s",cfg->port);
		else if (!(ld = ldap_init(cfg->host,port)))
			ldapauth_warn(r,"Internal error: %s",strerror(errno));
		else
			ldapauth_debug(r,"Opened LDAP context to host %s/%d",cfg->host,(int)port);

		if (ld && LDAP_OPT_SUCCESS != ldap_set_option(ld,LDAP_OPT_PROTOCOL_VERSION,&version)) {
			ldapauth_warn(r,"Failed to set protocol version to %s",cfg->version);

			ldap_unbind_s(ld);

			ld	= 0;
		}

		if (ld && 0 == strcasecmp(cfg->tls_mode,"StartTLS")) {
			if (LDAP_OPT_SUCCESS != ldap_start_tls_s(ld,NULL,NULL)) {
				ldapauth_warn(r,"StartTLS failed");

				ldap_unbind_s(ld);

				ld	= 0;
			} else
				ldapauth_debug(r,"Set StartTLS mode");
		}

		ldapauth_debug(r,"@line %d",__LINE__);

	} else {
		ldapauth_warn(r,"Invalid AuthLDAPMode: %s",cfg->tls_mode);
	}

	return ld;
} /* ldapauth_open() */


static int ldapauth_bind(request_rec *r, struct auth_config *cfg, LDAP *ld, char *passwd) {
	int ret	= ldap_simple_bind_s(ld,nulstr(cfg->binddn),nulstr(passwd));

	if (ret != LDAP_SUCCESS) {
			ldapauth_warn(r,"Failed to bind with DN '%s': %s",nulstr(cfg->binddn),ldap_err2string(ret));

			return -1;
	}

	return 0;	
} /* ldapauth_bind() */


static int ldapauth_str2scope(request_rec *r, const char *scope) {
	if (0 == strcasecmp(scope,"base")) {
		return LDAP_SCOPE_BASE;
	} else if (0 == strcasecmp(scope,"sub") || 0 == strcasecmp(scope,"subtree")) {
		return LDAP_SCOPE_SUBTREE;
	} else if (0 == strcasecmp(scope,"one") || 0 == strcasecmp(scope,"onelevel")) {
		return LDAP_SCOPE_ONELEVEL;
	} else {
		if (r)
			ldapauth_warn(r,"Invalid scope: %s",scope);

		return LDAP_SCOPE_ONELEVEL;
	}
} /* ldapauth_str2scope() */


static int ldapauth_lderrno(LDAP *ld) {
	int err;

	if (LDAP_OPT_SUCCESS != ldap_get_option(ld,LDAP_OPT_ERROR_NUMBER,&err))
		return LDAP_LOCAL_ERROR;

	return err;
} /* ldapauth_lderrno() */


static char *ldapauth_finddn(request_rec *r, struct auth_config *cfg, LDAP *ld, char *user) {
	struct timeval tv	= { 15, 0 };
	char *attrs[]		= { "DN", 0 };
	char *filter		= 0;
	LDAPMessage *res	= 0;
	LDAPMessage *ent	= 0;	/* No need to free. */
	char *entdn		= 0;
	char *userdn		= 0;
	int scope, ret, nentries;

	if (!(filter = create_filter(r->pool,cfg,user))) {
		ldapauth_warn(r,"Internal error: %s",strerror(errno));

		return 0;
	}

	ldapauth_debug(r,"filter: %s",filter);
	ldapauth_debug(r,"scope: %s",cfg->scope);

	scope	= ldapauth_str2scope(r,cfg->scope);

	/* FIXME: Make timeout configurable. */
	ret	= ldap_search_st(ld,nulstr(cfg->basedn),scope,filter,attrs,0,&tv,&res);

	if (ret != LDAP_SUCCESS) {
		ldapauth_warn(r,"Internal error: %s",ldap_err2string(ret));
		
		return 0;
	}

	ret	= ldap_result2error(ld,res,0);

	if (ret != LDAP_SUCCESS) {
		ldapauth_warn(r,"search failed: %s",ldap_err2string(ret));

		if (ret == LDAP_RES_SEARCH_RESULT) {
			int err			= 0;
			char *dn		= 0;
			char *msg		= 0;
			char **refs		= 0;
			LDAPControl **ctls	= 0;

			if (LDAP_SUCCESS == ldap_parse_result(ld,res,&err,&dn,&msg,&refs,&ctls,0) && msg)
				ldapauth_warn(r,"server said: %s",msg);

			if (dn)
				ldap_memfree(dn);
			if (msg)
				ldap_memfree(msg);
			if (refs)
				ldap_value_free(refs);
			if (ctls)
				ldap_controls_free(ctls);
		}

		goto done;
	}

	nentries	= ldap_count_entries(ld,res);

	if (nentries < 0) {
		ldapauth_warn(r,"search failed: %s",ldap_err2string(ldapauth_lderrno(ld)));

		goto done;
	}

	ldapauth_debug(r,"Got %d entries in bind DN search",nentries);

	if (!nentries)
		goto done;

	if (nentries > 1) {
		ldapauth_warn(r,"search failed: Got %d entries in bind DN",nentries);

		goto done;
	}

	if (!(ent = ldap_first_entry(ld,res))) {
		ldapauth_warn(r,"search failed: %s",ldap_err2string(ldapauth_lderrno(ld)));

		goto done;
	}

	if (!(entdn = ldap_get_dn(ld,ent))) {
		ldapauth_warn(r,"search failed: %s",ldap_err2string(ldapauth_lderrno(ld)));

		goto done;
	}

	userdn	= ap_pstrdup(r->pool,entdn);

	ldap_memfree(entdn);

	if (!userdn) {
		ldapauth_warn(r,"search failed: %s",strerror(errno));
	}

done:
	if (res)
		ldap_msgfree(res);

	return userdn;	
} /* ldapauth_finddn() */


static int ldapauth_auth(request_rec *r) {
	struct auth_config *auth_cfg;
	const char *auth_type, *auth_user, *auth_pass, *auth_userdn;
	LDAP *ld;
	int ret;

	auth_type	= (char *)ap_auth_type(r);

	if (!auth_type || 0 != strcasecmp(auth_type,"Basic")) {
		ap_note_basic_auth_failure(r);

		ldapauth_warn(r,"Only `Basic' authentication supported");

		return HTTP_INTERNAL_SERVER_ERROR;
	}

	if (OK != (ret = ap_get_basic_auth_pw(r,&auth_pass)))
		return ret;

	auth_user	= r->connection->user;

	if (!auth_user || !*auth_user) {
		ap_note_basic_auth_failure(r);

		ldapauth_warn(r,"No username provided");

		return HTTP_UNAUTHORIZED;
	}

	ldapauth_debug(r,"Checking user %s",auth_user);

	auth_cfg	= ap_get_module_config(r->per_dir_config,&openldap_auth_module);

	if (!auth_cfg)
		return HTTP_INTERNAL_SERVER_ERROR;

	ldapauth_debug(r,"Read config");

	if (!auth_cfg->enabled) {
		ldapauth_debug(r,"AuthLDAP is Off");

		return DECLINED;
	}

	ldapauth_debug(r,"Opening connection to LDAP");

	if (!(ld = ldapauth_open(r,auth_cfg)))
		return HTTP_INTERNAL_SERVER_ERROR;

	ldapauth_debug(r,"@line %d",__LINE__);

	if (auth_cfg->binddn && *auth_cfg->binddn) {
		if (0 != ldapauth_bind(r,auth_cfg,ld,auth_cfg->password)) {
			ldap_unbind_s(ld);

			return HTTP_INTERNAL_SERVER_ERROR;
		}
	}

	ldapauth_debug(r,"@line %d",__LINE__);

	if (!(auth_userdn = ldapauth_finddn(r,auth_cfg,ld,(char *)auth_user))) {
		ldap_unbind_s(ld);

		return HTTP_UNAUTHORIZED;
	}

	ldapauth_debug(r,"Got user DN '%s'",auth_userdn);

	ret	= ldap_simple_bind_s(ld,nulstr(auth_userdn),nulstr(auth_pass));

	ldap_unbind_s(ld);

	if (ret != LDAP_SUCCESS) {
		ldapauth_warn(r,"Bind failed for user '%s': %s",auth_user,ldap_err2string(ret));

		return HTTP_UNAUTHORIZED;
	} else
		return OK;
} /* ldapauth_auth() */


static void ldapauth_init(server_rec *s, pool *p) {
	int nocerts	= LDAP_OPT_X_TLS_ALLOW;

	if (LDAP_OPT_SUCCESS != ldap_set_option(NULL,LDAP_OPT_X_TLS_REQUIRE_CERT,&nocerts))
		ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,s,"[AuthLDAP] ldap_set_option(LDAP_OPT_X_TLS) failed");

	return /* void */;
} /* ldapauth_init() */


/*
 * Apache config command table
 */
static const command_rec ctable[] = {
	{
		"AuthLDAP",
		ap_set_flag_slot,
		(void *)XtOffsetOf(struct auth_config,enabled),
		OR_AUTHCFG,
		FLAG,
		"Toggle LDAP authentication: On | Off"
	},{
		"AuthLDAPRequireSSL",
		ap_set_flag_slot,
		(void *)XtOffsetOf(struct auth_config,require_ssl),
		OR_AUTHCFG,
		FLAG,
		"Toggle secure connection requirement: On | Off"
	},{
		"AuthLDAPHost",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,host),
		OR_AUTHCFG,
		TAKE1,
		"LDAP host: <domain|ip>[ <domain|ip>[ ...]]"
	},{
		"AuthLDAPPort",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,port),
		OR_AUTHCFG,
		TAKE1,
		"LDAP port: <1-65535>"
	},{
		"AuthLDAPVersion",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,version),
		OR_AUTHCFG,
		TAKE1,
		"LDAP Version: <2|3>"
	},{
		"AuthLDAPMode",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,tls_mode),
		OR_AUTHCFG,
		TAKE1,
		"LDAP most: LDAPS|StartTLS|Plain"
	},{
		"AuthLDAPBindDN",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,binddn),
		OR_AUTHCFG,
		TAKE1,
		"LDAP bind DN"
	},{
		"AuthLDAPBindPass",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,password),
		OR_AUTHCFG,
		TAKE1,
		"LDAP bind password"
	},{
		"AuthLDAPBaseDN",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,basedn),
		OR_AUTHCFG,
		TAKE1,
		"LDAP base DN"
	},{
		"AuthLDAPScope",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,scope),
		OR_AUTHCFG,
		TAKE1,
		"LDAP scope: sub[tree]|one|base"
	},{
		"AuthLDAPFilter",
		ap_set_string_slot,
		(void *)XtOffsetOf(struct auth_config,filter),
		OR_AUTHCFG,
		TAKE1,
		"LDAP filter"
	},{ NULL }
};


/*
 * Apache module parameter definitions
 *
 * NOTE:
 * 	[N]: N is the order in which the routine is called, if called
 */
module MODULE_VAR_EXPORT openldap_auth_module = {
	STANDARD_MODULE_STUFF,
	NULL,			/* module initializer */
	create_dir_config,	/* per-directory config creator */
	NULL,			/* directory config merger */
	NULL,			/* server config creator */
	NULL,			/* server config merger */
	ctable,			/* command table */
	NULL,			/* [9] list of handlers */
	NULL,			/* [2] filename-to-URI translation */
	ldapauth_auth,		/* [5] check/validate user_id */
	NULL,			/* [6] check user_id is valid here */
	NULL,			/* [4] check access by host address */
	NULL,			/* [7] MIME type checker/settler */
	NULL,			/* [8] fixups */
	NULL,			/* [10] logger */
	NULL,			/* [3] header parser */
	ldapauth_init,		/* child_init */
	NULL,			/* child_exit */
	NULL,			/* [1] post read-request */
};

